/*
    Pacman Instructional Emulator
    User interface for Windows
    Copyright (c) 1997-2003,2004 Alessandro Scotti
    User interface for BeOS based on Alessandro Scotti's work by Carwyn Jones, (2004)
    http://www.ascotti.org/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <Application.h>
#include <SoundPlayer.h>
#include <Roster.h>
#include <MenuItem.h>
#include <MenuBar.h>
#include <Entry.h>
#include <Path.h>
#include <Window.h>
#include <Alert.h>
#include <unistd.h>
#include <stdio.h>
#include <assert.h>

#include "pie_key.h"
#include "resource.h"

#include "arcade.h"
#include "dib24_beos.h"

// BeOS specific classes, defines

#define MF_CHECKED true
#define MF_UNCHECKED false
enum { MB_OK, MB_ICONINFORMATION, MB_ICONERROR };

class pac_window : public BWindow
{
public:
	pac_window(BRect rect);
	~pac_window();
	void ResizeTo(float width, float height);
	void MessageReceived(BMessage *msg);
	bool QuitRequested();
	uint32 AddMenu();
	BMenuBar *Bar() const;
	static int32 pac_thread(void *object);
	int32 pac_thread();
private:
	BMenuBar *pac_bar;
	BView *pac_view;
	thread_id pac_id;
	volatile bool quitting;
	uint32 menu_height;
};

class pac_application : public BApplication
{
public:
	pac_application();
};

class pac_sound : public BSoundPlayer
{
public:
	pac_sound(media_raw_audio_format *format);
	static void callback(void *cookie, void *buffer, size_t len, const media_raw_audio_format &format);
	void SyncWrite(uint8 *data, size_t len);
private:
	uint8 *mBuffer;
	size_t mWrite, mPlay, mTotal;
};
// End BeOS classes

unsigned framesPerSecond    = PacmanMachine::VideoFrequency;
unsigned samplesPerSecond   = 44100;    // Valid values: 11025, 22050, 44100
unsigned bytesPerSample     = 2;        // Valid values: 1 or 2
unsigned samplesPerFrame    = samplesPerSecond / framesPerSecond;
unsigned sampleBytesPerFrame= bytesPerSample * samplesPerFrame;
unsigned soundBuffers       = 3;

static pac_window  *hWindow = NULL;         // Handle of main window
static int      windowWidth = PacmanMachine::ScreenWidth; // Width of client area
static int      windowHeight = PacmanMachine::ScreenHeight; // Height of client area
static bool     boIsActive = FALSE;     // Program active/inactive
static char     szAppName[] = "PacMan";

// Bitmap used for off-screen rendering
static Dib24    *dCanvas = 0;

// DIP switches
static unsigned dipDifficulty   = PacmanMachine::DipDifficulty_Normal;
static unsigned dipGhostNames   = PacmanMachine::DipGhostNames_Normal;
static unsigned dipLivesPerGame = PacmanMachine::DipLives_3;   
static unsigned dipCabinet      = PacmanMachine::DipCabinet_Upright;   
static unsigned dipBonus        = PacmanMachine::DipBonus_10000;

static pac_sound *pie_player = 0; // Main BeOS sound object

static char szHelp[] =
    "1 = one player\n"
    "2 = two players\n"
    "3 = insert coin in slot 1\n"
    "4 = insert coin in slot 2\n"
    "5 = add credit\n"
    "A = toggle rack advance switch\n"
    "F = activate speed hack\n"
    "S = deactivate speed hack\n"
    "T = toggle test switch\n"
    "Left/Up/Right/Down arrow keys = move joystick\n"
    "F2 = single/double window size\n"
    "F5/F6/F7 = save game snapshot\n"
    "F9 = restore F5 snapshot\n"
    "Escape = quit";

static char szAbout[] = 
    "PIE - Pacman Instructional Emulator\n"
    "Copyright (c) 1997-2003,2004 Alessandro Scotti\n\n"
    "BeOS port by Carwyn Jones, 2004\n"
    "\n"
    "http://www.ascotti.org/";

const char * SnapF5 = "snap.f5";
const char * SnapF6 = "snap.f6";
const char * SnapF7 = "snap.f7";

enum {
    GameNone = 0,
    GamePacmanMidway,
    GamePacmanNamco,
    GameMsPacman
};

PacmanMachine * machine = 0; // Arcade emulator
int gameLoaded = GameNone; // What type of machine is currently loaded

int * rawSoundBuf = 0; // Raw sound buffer used for machine->playSound()
bool applicationInitialized = FALSE;

void MessageBox(int no, const char *text, const char *title, int type)
{
	(new BAlert("text", text, "Ok", NULL, NULL, B_WIDTH_FROM_LABEL))->Go(0);
}

void terminateDirectDraw()
{
	delete dCanvas;
}

bool initializeDirectDraw( BWindow *hWnd )
{
    return true;
}

bool showError( const char * msg )
{
    MessageBox( 0, msg, "Error", MB_OK | MB_ICONERROR );
    return FALSE;
}

bool loadFile( const char * name, unsigned char * buf, int len, bool showErrorMsg = FALSE )
{
    char msg[200];
    FILE * f = fopen( name, "rb" );

    if( f == NULL ) {
        sprintf( msg, "Cannot open '%s'", name );
        if( showErrorMsg ) showError( msg );
        return FALSE;
    }
    
    int n = fread( buf, len, 1, f );
    
    fclose( f );

    if( n != 1 ) {
        sprintf( msg, "Error reading from '%s'", name );
        if( showErrorMsg ) showError( msg );
        return FALSE;
    }

    return TRUE;
}

int menuCheck( int condition )
{
    return condition ? MF_CHECKED : MF_UNCHECKED;
}

/*
    Sets the game DIP switches and updates the program menu accordingly.
*/

BMenuBar *GetMenu(pac_window *window)
{
	return window->Bar();
}

void CheckMenuItem(BMenu *menu, uint32 msg, bool value)
{
	BMenuItem *item = menu->FindItem(msg);
	if(item) item->SetMarked(value);
}

void setDipSwitches()
{
    BMenu   *hMenu = GetMenu( hWindow );

    CheckMenuItem( hMenu, IDM_NORMAL, menuCheck(dipDifficulty == PacmanMachine::DipDifficulty_Normal) );
    CheckMenuItem( hMenu, IDM_HARD, menuCheck(dipDifficulty == PacmanMachine::DipDifficulty_Hard) );
  
    CheckMenuItem( hMenu, IDM_LIFE_0, menuCheck(dipLivesPerGame == PacmanMachine::DipLives_1) );
    CheckMenuItem( hMenu, IDM_LIFE_1, menuCheck(dipLivesPerGame == PacmanMachine::DipLives_2) );
    CheckMenuItem( hMenu, IDM_LIFE_2, menuCheck(dipLivesPerGame == PacmanMachine::DipLives_3) );
    CheckMenuItem( hMenu, IDM_LIFE_3, menuCheck(dipLivesPerGame == PacmanMachine::DipLives_5) );

    CheckMenuItem( hMenu, IDM_GHOST_0, menuCheck(dipGhostNames == PacmanMachine::DipGhostNames_Normal) );
    CheckMenuItem( hMenu, IDM_GHOST_1, menuCheck(dipGhostNames == PacmanMachine::DipGhostNames_Alternate) );

    CheckMenuItem( hMenu, IDM_CABINET_UPRIGHT, menuCheck(dipCabinet == PacmanMachine::DipCabinet_Upright) );
    CheckMenuItem( hMenu, IDM_CABINET_COCKTAIL, menuCheck(dipCabinet == PacmanMachine::DipCabinet_Cocktail) );

    CheckMenuItem( hMenu, IDM_BONUS_10000, menuCheck(dipBonus == PacmanMachine::DipBonus_10000) );
    CheckMenuItem( hMenu, IDM_BONUS_15000, menuCheck(dipBonus == PacmanMachine::DipBonus_15000) );
    CheckMenuItem( hMenu, IDM_BONUS_20000, menuCheck(dipBonus == PacmanMachine::DipBonus_20000) );
    CheckMenuItem( hMenu, IDM_BONUS_NONE, menuCheck(dipBonus == PacmanMachine::DipBonus_None) );

	if(!machine) return;
    unsigned dip = machine->getDipSwitches() & (PacmanMachine::DipPlay_Mask | PacmanMachine::DipMode_Mask);
    machine->setDipSwitches( dip | dipDifficulty | dipGhostNames | dipLivesPerGame | dipBonus | dipCabinet );
}

/*
    Saves a snapshot of the game.
*/
void takeSnapshot( const char * filename )
{
    unsigned bufsize = 1 + machine->getSizeOfSnapshotBuffer();
    unsigned char * buffer = new unsigned char [ bufsize ];

    *buffer = (unsigned char)(gameLoaded);
    machine->takeSnapshot( buffer+1 );

    FILE * f = fopen( filename, "wb" );

    assert( f != NULL );

    fwrite( buffer, 1, bufsize, f );
    fclose( f );

    delete buffer;
}

/*
    Restores a snapshot of the game.
*/
void restoreSnapshot( const char * filename )
{
    PacmanMachine * newmachine = 0;

    FILE * f = fopen( filename, "rb" );

    if( f != 0 ) {
        unsigned char game;

        fread( &game, 1, 1, f );

        switch( game ) {
        case GamePacmanMidway:
        case GamePacmanNamco:
            newmachine = new PacmanMachine;
            break;
        case GameMsPacman:
            newmachine = new MsPacmanMachine;
            break;
        }

        if( newmachine != 0 ) {
            unsigned bufsize = newmachine->getSizeOfSnapshotBuffer();
            unsigned char * buffer = new unsigned char [ bufsize ];

            if( fread( buffer, 1, bufsize, f ) == bufsize ) {
                newmachine->restoreSnapshot( buffer );

                dipDifficulty = newmachine->getDipSwitches() & PacmanMachine::DipDifficulty_Mask;
                dipGhostNames = newmachine->getDipSwitches() & PacmanMachine::DipGhostNames_Mask;
                dipLivesPerGame = newmachine->getDipSwitches() & PacmanMachine::DipLives_Mask;
                dipBonus = newmachine->getDipSwitches() & PacmanMachine::DipBonus_Mask;
                setDipSwitches();

                delete machine;
                machine = newmachine;
            }
            else {
                delete newmachine;
            }

            fclose( f );

            delete buffer;
        }
    }
}

/*
    Initializes the sound library and creates the DirectX sound objects
    used by the application.
*/
bool initSound( void )
{
	media_raw_audio_format format = {
		samplesPerSecond,
		1,
		media_raw_audio_format::B_AUDIO_SHORT,
		B_MEDIA_LITTLE_ENDIAN,
		sampleBytesPerFrame
	};
	
	pie_player = new pac_sound(&format);
	if(pie_player->InitCheck() != B_OK) {
		delete pie_player; pie_player = 0;
		return false;
	}
	
	pie_player->Start();
	pie_player->SetHasData(true);
	
	return true;
}

/*
    Activates/deactivates the program.

    Note: when the program is not active, emulation and sound playing are stopped.
*/
void setActive( bool active )
{
    if( !applicationInitialized || (gameLoaded == GameNone) )
        return;

    active = active ? TRUE : FALSE; // Normalize the boolean values!

    if( active != boIsActive ) {
        boIsActive = active;
        if( active ) {
        }
        else {
        }
    }
}

bool loadGameROM( int game )
{
    bool romLoaded = FALSE;
    PacmanMachine * newmachine = 0;

    unsigned char   buffer[ 0x6000 ];
    unsigned char   palette[0x20];
    unsigned char   color[0x100];

    // Files common to all ROM sets
    if( (! loadFile( "82s126.4a", color, sizeof(color), TRUE)) ||
        (! loadFile( "82s123.7f", palette, sizeof(palette), TRUE)) )
    {
        return FALSE;
    }

    // Basic game ROMs
    switch( game ) {
    case GamePacmanNamco:
        newmachine = new PacmanMachine;
        romLoaded = loadFile( "namcopac.6e", buffer, 0x1000 ) &&
            loadFile( "namcopac.6f", buffer+0x1000, 0x1000 ) &&
            loadFile( "namcopac.6h", buffer+0x2000, 0x1000 ) &&
            loadFile( "namcopac.6j", buffer+0x3000, 0x1000 ) &&
            loadFile( "pacman.5e", buffer+0x4000, 0x1000 ) &&
            loadFile( "pacman.5f", buffer+0x5000, 0x1000 );
        break;
    case GameMsPacman:
        newmachine = new MsPacmanMachine;
        romLoaded = loadFile( "pacman.6e", buffer, 0x1000 ) &&
            loadFile( "pacman.6f", buffer+0x1000, 0x1000 ) &&
            loadFile( "pacman.6h", buffer+0x2000, 0x1000 ) &&
            loadFile( "pacman.6j", buffer+0x3000, 0x1000 ) &&
            loadFile( "5e", buffer+0x4000, 0x1000 ) &&
            loadFile( "5f", buffer+0x5000, 0x1000 );
        break;
    case GamePacmanMidway:
        newmachine = new PacmanMachine;
        romLoaded = loadFile( "pacman.6e", buffer, 0x1000 ) &&
            loadFile( "pacman.6f", buffer+0x1000, 0x1000 ) &&
            loadFile( "pacman.6h", buffer+0x2000, 0x1000 ) &&
            loadFile( "pacman.6j", buffer+0x3000, 0x1000 ) &&
            loadFile( "pacman.5e", buffer+0x4000, 0x1000 ) &&
            loadFile( "pacman.5f", buffer+0x5000, 0x1000 );
        break;
    }

    // Set ROMs and, if necessary, load aux ROMs
    if( romLoaded ) {
        newmachine->setROM( buffer );
        newmachine->setVideoROMs( buffer+0x4000, buffer+0x5000 );
        newmachine->setColorROMs( palette, color );

        // Load Ms. Pacman ROMs (must do it after the basic ROMs!)
        if( game == GameMsPacman ) {
            romLoaded = loadFile( "u5", buffer+0x0000, 0x0800 ) &&
                loadFile( "u6", buffer+0x0800, 0x1000 ) &&
                loadFile( "u7", buffer+0x1800, 0x1000 );

            // Add extra (ROM) board to emulator
            reinterpret_cast<MsPacmanMachine *>(newmachine)->setAuxROM( buffer+0x0000, buffer+0x0800, buffer+0x1800 );
        }
    }

    if( romLoaded ) {
        delete machine;
        machine = newmachine;
        machine->reset();
        gameLoaded = game;
    }
    else {
        delete newmachine;
    }

    return romLoaded;
}

void loadGame( int game, int quiet )
{
    if( ! loadGameROM( game ) ) {
        if( ! quiet ) showError( "Game ROMs not found!" );
    }

    setActive( TRUE );
}

void blitGameScreen() {}

void drawGameScreen() {}

/*
    Initializes the game engine.
*/
bool gameInit( void )
{
    // Initialize the sound library
    if( ! initSound() )
        return FALSE;

    if( ! initializeDirectDraw(hWindow) )
        return FALSE;

    dCanvas->clear();
    drawGameScreen();

    rawSoundBuf = new int[ samplesPerFrame ];

    // Load game ROMs
    loadGame( GamePacmanMidway, TRUE );

    applicationInitialized = TRUE;

    return TRUE;
}

/*
    Runs the game engine for one frame.
*/
void gameProc( void )
{
    if( ! boIsActive )
        return;

    // Run the machine for a frame
    machine->run();

    bool drawVideo = (machine->getFrameCount() & 1) != 0;

    // Prepare the sound buffer
    machine->playSound( rawSoundBuf, samplesPerFrame, samplesPerSecond );

    // Draw the video image into the off-screen bitmap
    if( drawVideo ) {
        unsigned char video_buffer[ PacmanMachine::ScreenWidth * PacmanMachine::ScreenHeight ];
        unsigned char * vbuf = video_buffer;
        const unsigned * palette = machine->getPalette();
        
        machine->renderVideo( vbuf );

        for( int y=0; y<PacmanMachine::ScreenHeight; y++ ) {
            // Use direct access to the bitmap scanline in place of
            // setFastPixel() for performance (the time saved here comes useful when
            // the window is zoomed, which slows down painting considerably).
            unsigned char * dst = dCanvas->getScanLine( y );
            for( int x=0; x<PacmanMachine::ScreenWidth; x++ ) {
                unsigned color = palette[ *vbuf++ ];
                
                *dst++ = (unsigned char) (color >> 16);
                *dst++ = (unsigned char) (color >> 8);
                *dst++ = (unsigned char) (color);
                dst++;
            }
        }
    }
    
    uint8 *pwBuf = new uint8[sampleBytesPerFrame];
    uint8 *sndBuf = (uint8 *)pwBuf;
    int *rawBuf = rawSoundBuf;
    
	for( int i=0; i<(int)samplesPerFrame; i++ ) {
		short w = (short)(*rawBuf++ * 85);
		*sndBuf++ = (uint8) (w);
		*sndBuf++ = (uint8) (w >> 8);
	}
	
	pie_player->SyncWrite(pwBuf, sampleBytesPerFrame);
	
	delete [] pwBuf;
	
    // Blit last frame and draw current in the back buffer
    if( drawVideo ) {
        blitGameScreen();
        drawGameScreen();
    }
}

/*
    Terminates the game engine.
*/
void gameTerm()
{
	if(pie_player) {
		pie_player->Stop();
		delete pie_player; pie_player = 0;
	}
    terminateDirectDraw();
    delete rawSoundBuf;
}

void setWindowSize( int sizeFactor )
{
    windowWidth = sizeFactor*PacmanMachine::ScreenWidth; 
    windowHeight = sizeFactor*PacmanMachine::ScreenHeight;
    hWindow->ResizeTo(windowWidth - 1, windowHeight - 1);
}

/*
    Window procedure.
*/

pac_window::pac_window(BRect rect)
	: BWindow(rect, "Pie", B_TITLED_WINDOW, 0),
	quitting(false)
{
	rect.OffsetTo(B_ORIGIN);
	menu_height = AddMenu();
	rect.top = (float)menu_height + 1;
	
	pac_view = new BView(rect, "pac_view", B_FOLLOW_ALL, B_WILL_DRAW);
	pac_view->SetViewColor(0, 0, 0);
	AddChild(pac_view);
	
	dCanvas = new Dib24( PacmanMachine::ScreenWidth, PacmanMachine::ScreenHeight );
	
	ResizeTo(windowWidth - 1, windowHeight - 1);
	pac_id = spawn_thread(pac_thread, "pie", B_NORMAL_PRIORITY, this);
	resume_thread(pac_id);
}

void pac_window::ResizeTo(float width, float height)
{
	height += menu_height + 1;
	BWindow::ResizeTo(width, height);
}

uint32 pac_window::AddMenu()
{
	pac_bar = new BMenuBar(BRect(0, 0, 0, 0), "bar");
	BMenuItem *item = 0;
	
	BMenu *file = new BMenu("File"); // File menu
	file->AddItem(new BMenuItem("Load Midway Pacman", new BMessage(IDM_LOAD_PACMAN)));
	file->AddItem(new BMenuItem("Load Namco Pacman", new BMessage(IDM_LOAD_NAMCOPAC)));
	file->AddItem(new BMenuItem("Load Ms. Pacman", new BMessage(IDM_LOAD_MSPACMAN)));
	file->AddSeparatorItem();
	file->AddItem(new BMenuItem("Restart", new BMessage(IDM_NEW)));
	file->AddSeparatorItem();
	file->AddItem(new BMenuItem("Restore F5 Snapshot", new BMessage(IDM_SNAP_0)));
	file->AddItem(new BMenuItem("Restore F6 Snapshot", new BMessage(IDM_SNAP_1)));
	file->AddItem(new BMenuItem("Restore F7 Snapshot", new BMessage(IDM_SNAP_2)));
	file->AddSeparatorItem();
	file->AddItem(new BMenuItem("Exit", new BMessage(IDM_EXIT)));
	pac_bar->AddItem(file);
	
	BMenu *options = new BMenu("Options"); // Options menu
	options->AddItem(new BMenuItem("Toggle window size", new BMessage(IDM_WINDOW_SIZE)));
	options->AddSeparatorItem();
	options->AddItem(item = new BMenuItem("Normal", new BMessage(IDM_NORMAL)));
	item->SetMarked(true);
	options->AddItem(new BMenuItem("Hard(more difficult)", new BMessage(IDM_HARD)));
	options->AddSeparatorItem();
	options->AddItem(new BMenuItem("1 Pacman per Game", new BMessage(IDM_LIFE_0)));
    options->AddItem(new BMenuItem("2 Pacmen per Game", new BMessage(IDM_LIFE_1)));
    options->AddItem(item = new BMenuItem("3 Pacmen per Game", new BMessage(IDM_LIFE_2)));
    item->SetMarked(true);
    options->AddItem(new BMenuItem("5 Pacmen per Game",  new BMessage(IDM_LIFE_3)));
	options->AddSeparatorItem();
    options->AddItem(item = new BMenuItem("Bonus pacman at 10000",	new BMessage(IDM_BONUS_10000)));
    item->SetMarked(true);
    options->AddItem(new BMenuItem("Bonus pacman at 15000",	new BMessage(IDM_BONUS_15000)));
    options->AddItem(new BMenuItem("Bonus pacman at 20000",	new BMessage(IDM_BONUS_20000)));
    options->AddItem(new BMenuItem("No bonus", new BMessage(IDM_BONUS_NONE)));
    options->AddSeparatorItem();
    options->AddItem(item = new BMenuItem("Normal Ghost Names", new BMessage(IDM_GHOST_0)));
    item->SetMarked(true);
    options->AddItem(new BMenuItem("Alternate Ghost Names", new BMessage(IDM_GHOST_1)));
    options->AddSeparatorItem();
    options->AddItem(item = new BMenuItem("Cabinet upright", new BMessage(IDM_CABINET_UPRIGHT)));
    item->SetMarked(true);
    options->AddItem(new BMenuItem("Cabinet cocktail", new BMessage(IDM_CABINET_COCKTAIL)));
	pac_bar->AddItem(options);
	
	BMenu *help = new BMenu("Help"); // Help menu
	help->AddItem(new BMenuItem("Keys", new BMessage(IDM_HELP_KEYS)));
	help->AddSeparatorItem();
    help->AddItem(new BMenuItem("About", new BMessage(IDM_ABOUT)));
	pac_bar->AddItem(help);
	AddChild(pac_bar);
	
	return (uint32)pac_bar->Bounds().Height();
}

BMenuBar *pac_window::Bar() const
{
	return pac_bar;
}

void pac_window::MessageReceived(BMessage *msg)
{
	int32 wParam;
	switch(msg->what) {
            case IDM_LOAD_PACMAN:
                loadGame( GamePacmanMidway, FALSE );
                break;
            case IDM_LOAD_NAMCOPAC:
                loadGame( GamePacmanNamco, FALSE );
                break;
            case IDM_LOAD_MSPACMAN:
                loadGame( GameMsPacman, FALSE );
                break;
            case IDM_NEW:
                if(machine) machine->reset();
                break;
            case IDM_SNAP_0:
                restoreSnapshot( SnapF5 );
                break;
            case IDM_SNAP_1:
                restoreSnapshot( SnapF6 );
                break;
            case IDM_SNAP_2:
                restoreSnapshot( SnapF7 );
                break;
            case IDM_EXIT:
                PostMessage(B_QUIT_REQUESTED);
                break;
            case IDM_WINDOW_SIZE:
                setWindowSize( (windowWidth == PacmanMachine::ScreenWidth) ? 2 : 1 );
                break;
            case IDM_HELP_KEYS:
                MessageBox( 0, szHelp, "Keys", MB_OK | MB_ICONINFORMATION );
                break;
            case IDM_ABOUT:
                MessageBox( 0, szAbout, szAppName, MB_OK | MB_ICONINFORMATION );
                break;
            case IDM_LIFE_0:
                dipLivesPerGame = PacmanMachine::DipLives_1;
                setDipSwitches();
                break;
            case IDM_LIFE_1:
                dipLivesPerGame = PacmanMachine::DipLives_2;
                setDipSwitches();
                break;
            case IDM_LIFE_2:
                dipLivesPerGame = PacmanMachine::DipLives_3;
                setDipSwitches();
                break;
            case IDM_LIFE_3:
                dipLivesPerGame = PacmanMachine::DipLives_5;
                setDipSwitches();
                break;
            case IDM_HARD:
                dipDifficulty = PacmanMachine::DipDifficulty_Hard;
                setDipSwitches();
                break;
            case IDM_NORMAL:
                dipDifficulty = PacmanMachine::DipDifficulty_Normal;
                setDipSwitches();
                break;
            case IDM_GHOST_0:
                dipGhostNames = PacmanMachine::DipGhostNames_Normal;
                setDipSwitches();
                break;
            case IDM_GHOST_1:
                dipGhostNames = PacmanMachine::DipGhostNames_Alternate;
                setDipSwitches();
                break;
            case IDM_BONUS_10000:
                dipBonus = PacmanMachine::DipBonus_10000;
                setDipSwitches();
                break;
            case IDM_BONUS_15000:
                dipBonus = PacmanMachine::DipBonus_15000;
                setDipSwitches();
                break;
            case IDM_BONUS_20000:
                dipBonus = PacmanMachine::DipBonus_20000;
                setDipSwitches();
                break;
            case IDM_BONUS_NONE:
                dipBonus = PacmanMachine::DipBonus_None;
                setDipSwitches();
                break;
            case IDM_CABINET_UPRIGHT:
                dipCabinet = PacmanMachine::DipCabinet_Upright;
                setDipSwitches();
                break;
            case IDM_CABINET_COCKTAIL:
                dipCabinet = PacmanMachine::DipCabinet_Cocktail;
                setDipSwitches();
                break;
			// Keyboard events
        case B_KEY_DOWN:
        case B_KEY_UP:
			if(msg->FindInt32("key", &wParam) != B_OK) break;
            // Handle machine device events first...
            {
                PacmanMachine::InputDeviceMode mode = (msg->what == B_KEY_DOWN) ? PacmanMachine::DeviceOn : PacmanMachine::DeviceOff;
                
                switch( wParam ) {
                case PIE_KEY1:
                    if(machine) machine->setDeviceMode( PacmanMachine::Key_OnePlayer, mode );
                    break;
                case PIE_KEY2:
                    if(machine) machine->setDeviceMode( PacmanMachine::Key_TwoPlayers, mode );
                    break;
                case PIE_KEY3:
                    if(machine) machine->setDeviceMode( PacmanMachine::CoinSlot_1, mode );
                    break;
                case PIE_KEY4:
                    if(machine) machine->setDeviceMode( PacmanMachine::CoinSlot_2, mode );
                    break;
                case PIE_KEY5:
                    if(machine) machine->setDeviceMode( PacmanMachine::Switch_AddCredit, mode );
                    break;
                case PIE_LEFT:
                    if(machine) machine->setDeviceMode( PacmanMachine::Joy1_Left, mode );
                    if(machine) machine->setDeviceMode( PacmanMachine::Joy2_Left, mode );
                    break;
                case PIE_RIGHT:
                    if(machine) machine->setDeviceMode( PacmanMachine::Joy1_Right, mode );
                    if(machine) machine->setDeviceMode( PacmanMachine::Joy2_Right, mode );
                    break;
                case PIE_UP:
                    if(machine) machine->setDeviceMode( PacmanMachine::Joy1_Up, mode );
                    if(machine) machine->setDeviceMode( PacmanMachine::Joy2_Up, mode );
                    break;
                case PIE_DOWN:
                    if(machine) machine->setDeviceMode( PacmanMachine::Joy1_Down, mode );
                    if(machine) machine->setDeviceMode( PacmanMachine::Joy2_Down, mode );
                    break;
                }
            }
            // ...then handle application events
            if( msg->what == B_KEY_DOWN ) switch( wParam ) {
            case PIE_A:
                if(machine) machine->setDeviceMode( PacmanMachine::Switch_RackAdvance, PacmanMachine::DeviceToggle );
                break;
            case PIE_F:
                if(machine) machine->setSpeedHack( 1 ); // Fast pacman
                break;
            case PIE_S:
                if(machine) machine->setSpeedHack( 0 ); // Slow (normal) pacman
                break;
            case PIE_T:
                if(machine) machine->setDeviceMode( PacmanMachine::Switch_Test, PacmanMachine::DeviceToggle );
                break;
            case PIE_SPACE:
                setActive( ! boIsActive );
                break;
            case PIE_ESCAPE:
                PostMessage( B_QUIT_REQUESTED );
                break;
            case PIE_F2:
                setWindowSize( (windowWidth == PacmanMachine::ScreenWidth) ? 2 : 1 );
                break;
            case PIE_F4:
                // Cheat: add one life
                {
                	if(!machine) break;
                    unsigned char * ram = const_cast<unsigned char *>(machine->getRAM());
                    ram[0x4E14]++;
                    ram[0x4E15]++;
                }
                break;
            case PIE_F5:
                takeSnapshot( SnapF5 );
                break;
            case PIE_F6:
                takeSnapshot( SnapF6 );
                break;
            case PIE_F7:
                takeSnapshot( SnapF7 );
                break;
            case PIE_F9:
                restoreSnapshot( SnapF5 );
                break;
            }
            break;
            default:
            	BWindow::MessageReceived(msg);
            	break;
            }
}

int32 pac_window::pac_thread(void *object)
{
	pac_window *pac = (pac_window *)object;
	return pac->pac_thread();
}

int32 pac_window::pac_thread()
{
	if(!gameInit()) return 0;
	setActive( TRUE );
	
	while(!quitting) {
		if( boIsActive ) {
			gameProc();
			if(Lock()) {
				pac_view->DrawBitmap(dCanvas->getHandle(), pac_view->Bounds());
				Unlock();
			}
		} else
			snooze(16000);
	}
	gameTerm();
	return 0;
}

bool pac_window::QuitRequested()
{
	be_app->PostMessage(B_QUIT_REQUESTED);
	return true;
}

pac_window::~pac_window()
{
	quitting = true;
	int32 count_locks = 0, pie;
	
	while(IsLocked()) {
		count_locks++;
		Unlock();
	}
	wait_for_thread(pac_id, &pie);
}

pac_sound::pac_sound(media_raw_audio_format *format)
	: BSoundPlayer(format, "pie"),
	mBuffer(0),
	mWrite(0),
	mPlay(0),
	mTotal(format->buffer_size * 8)
{
	SetBufferPlayer(callback);
	SetCookie(this);
	mBuffer = new uint8[mTotal];
	memset(mBuffer, 0, mTotal);
}

void pac_sound::callback(void *cookie, void *buffer, size_t len, const media_raw_audio_format &format)
{
	pac_sound *pie = (pac_sound *)cookie;
	
	size_t pos = pie->mPlay + len;
	size_t space = pie->mTotal - pie->mPlay;
	uint8 *out = (uint8 *)buffer;
	
	if(pos > pie->mTotal) {
		memcpy(out, &pie->mBuffer[pie->mPlay], space);
		out += space;
		len -= space;
		memcpy(out, &pie->mBuffer[0], len);
		pie->mPlay = pos - pie->mTotal;
	}
	else {
		memcpy(out, &pie->mBuffer[pie->mPlay], len);
		pie->mPlay = pos;
	}
}

void pac_sound::SyncWrite(uint8 *data, size_t len)
{
	while(mPlay == mWrite) snooze(1000);
	
	size_t pos = mWrite + len;
	size_t space = mTotal - mWrite;
	uint8 *out = (uint8 *)data;
	if(pos > mTotal) {
		memcpy(&mBuffer[mWrite], out, space);
		out += space;
		len -= space;
		memcpy(&mBuffer[0], out, len);
		mWrite = pos - mTotal;
	} else {
		memcpy(&mBuffer[mWrite], out, len);
		mWrite = pos;
	}
}

pac_application::pac_application()
	: BApplication("application/x-vnd.pie")
{
	app_info ai;
	BPath path;
	GetAppInfo(&ai);
	BEntry entry(&ai.ref);
	entry.GetPath(&path);
	path.GetParent(&path);
	chdir(path.Path());
	
	BRect r(200, 200, 400, 400);
	hWindow = new pac_window(r);
	hWindow->Show();
}

int main()
{
	pac_application app;
	app.Run();
	return 0;
}
